# 📫 MIME Message builder

`mime-msg-builder` is a Rust library for building MIME Messages.

The core feature of the library is the Rust implementation of the
[Emacs
MML](https://www.gnu.org/software/emacs/manual/html_node/emacs-mime/Composing.html)
module:

> Creating a MIME message is boring and non-trivial. Therefore, a
> library called mml has been defined that parses a language called
> MML (MIME Meta Language) and generates MIME messages.

```eml
From: alice@localhost
To: bob@localhost
Subject: MML simple

<#multipart type=alternative>
This is a plain text part.
<#part type=text/enriched>
<center>This is a centered enriched part</center>
<#/multipart>
```

compiles to:

```eml
Subject: MML simple
To: bob@localhost
From: alice@localhost
MIME-Version: 1.0
Date: Tue, 29 Nov 2022 13:07:01 +0000
Content-Type: multipart/alternative;
 boundary="4CV1Cnp7mXkDyvb55i77DcNSkKzB8HJzaIT84qZe"

--4CV1Cnp7mXkDyvb55i77DcNSkKzB8HJzaIT84qZe
Content-Type: text/plain; charset=utf-8
Content-Transfer-Encoding: 7bit

This is a plain text part.
--4CV1Cnp7mXkDyvb55i77DcNSkKzB8HJzaIT84qZe
Content-Type: text/enriched
Content-Transfer-Encoding: 7bit

<center>This is a centered enriched part</center>
--4CV1Cnp7mXkDyvb55i77DcNSkKzB8HJzaIT84qZe--
```

## Definition

From the
[documentation](https://www.gnu.org/software/emacs/manual/html_node/emacs-mime/MML-Definition.html):

> The MML language is very simple. It looks a bit like an SGML
> application, but it’s not.
> 
> The main concept of MML is the part. Each part can be of a different
> type or use a different charset. The way to delineate a part is with
> a ‘<#part ...>’ tag. Multipart parts can be introduced with the
> ‘<#multipart ...>’ tag. Parts are ended by the ‘<#/part>’ or
> ‘<#/multipart>’ tags. Parts started with the ‘<#part ...>’ tags are
> also closed by the next open tag.
> 
> […]
> 
> Each tag can contain zero or more parameters on the form
> ‘parameter=value’. The values may be enclosed in quotation marks,
> but that’s not necessary unless the value contains white space. So
> ‘filename=/home/user/#hello$^yes’ is perfectly valid.

## Examples

See
[`./examples`](https://git.sr.ht/~soywod/rust-mml/tree/master/item/examples):

```sh
cargo run --example
```

## Features

* Template builder
* Multipart support `<#multipart>…<#/multipart>`
* SinglePart support `<#part>…<#/part>`
* Attachment support `<#part filename=…>`
* Part encryption using custom system commands (default: `gpg
  --encrypt --armor --recipient <recipient> --quiet --output -`)
* Part signing using custom system commands (default: `gpg --sign
  --armor --quiet --output -`)

## Development

The development environment is managed by
[Nix](https://nixos.org/download.html). Running `nix-shell` will spawn
a shell with everything you need to get started with the lib: `cargo`,
`cargo-watch`, `rust-bin`, `rust-analyzer`…

```sh
# starts a nix shell
$ nix-shell

# builds the lib
$ cargo build
```

## Contributing

If you find a bug, feel free to open an issue at
[~soywod/mime-msg-builder](https://todo.sr.ht/~soywod/mime-msg-builder).

If you have a feature in mind, feel free to send a patchset at
https://git.sr.ht/~soywod/mime-msg-builder/send-email or use the
command `git send-email`.

You can also contact me directly on [Matrix](https://matrix.org/) at
[@soywod:matrix.org](https://matrix.to/#/@soywod:matrix.org).

## Sponsoring

[![GitHub](https://img.shields.io/badge/-GitHub%20Sponsors-fafbfc?logo=GitHub%20Sponsors&style=flat-square)](https://github.com/sponsors/soywod)
[![PayPal](https://img.shields.io/badge/-PayPal-0079c1?logo=PayPal&logoColor=ffffff&style=flat-square)](https://www.paypal.com/paypalme/soywod)
[![Ko-fi](https://img.shields.io/badge/-Ko--fi-ff5e5a?logo=Ko-fi&logoColor=ffffff&style=flat-square)](https://ko-fi.com/soywod)
[![Buy Me a Coffee](https://img.shields.io/badge/-Buy%20Me%20a%20Coffee-ffdd00?logo=Buy%20Me%20A%20Coffee&logoColor=000000&style=flat-square)](https://www.buymeacoffee.com/soywod)
[![Liberapay](https://img.shields.io/badge/-Liberapay-f6c915?logo=Liberapay&logoColor=222222&style=flat-square)](https://liberapay.com/soywod)
