//! Wrapper arround the [`std::process`] module for spawing processes
//! and redirecting standard input/output on all platforms.

use std::{
    env,
    io::{self, prelude::*},
    process::{Command, Stdio},
    result,
};
use thiserror::Error;

#[derive(Debug, Error)]
pub enum Error {
    #[error("cannot spawn process with command: {1}")]
    SpawnProcessError(#[source] io::Error, String),
    #[error("cannot get standard input")]
    GetStdinError,
    #[error("cannot write data to standard input")]
    WriteStdinError(#[source] io::Error),
    #[error("cannot get standard output")]
    GetStdoutError,
    #[error("cannot read data from standard output")]
    ReadStdoutError(#[source] io::Error),
}

type Result<T> = result::Result<T, Error>;

/// Runs the given command in a pipeline and returns the raw output
/// with the final exit code.
pub fn pipe(cmd: &str, input: &[u8]) -> Result<(Vec<u8>, usize)> {
    let mut output = Vec::new();

    let windows = cfg!(target_os = "windows")
        && env::var("MSYSTEM")
            .map(|env| !env.starts_with("MINGW"))
            .unwrap_or_default();

    let pipeline = if windows {
        Command::new("cmd")
            .args(&["/C", cmd])
            .stdin(Stdio::piped())
            .stdout(Stdio::piped())
            .stderr(Stdio::piped())
            .spawn()
    } else {
        Command::new("sh")
            .arg("-c")
            .arg(cmd)
            .stdin(Stdio::piped())
            .stdout(Stdio::piped())
            .stderr(Stdio::piped())
            .spawn()
    }
    .map_err(|err| Error::SpawnProcessError(err, cmd.to_string()))?;

    pipeline
        .stdin
        .ok_or_else(|| Error::GetStdinError)?
        .write_all(input)
        .map_err(Error::WriteStdinError)?;

    let code = pipeline
        .stdout
        .ok_or_else(|| Error::GetStdoutError)?
        .read_to_end(&mut output)
        .map_err(Error::ReadStdoutError)?;

    Ok((output, code))
}
